""""""

# Standard library modules.

# Third party modules.

# Local modules.
from pypenelopetools.penelope.separator import Separator

# Globals and constants variables.

INPUT_PHASE_SPACE_FILE = Separator('>>>>>>>> Input phase-space file (psf).')
GEOMETRY = Separator('>>>>>>>> Geometry and local simulation parameters.')
IMPACT_DETECTORS = Separator('>>>>>>>> Impact detectors (up to 25 different detectors).')
ABSORBED_DOSE_DISTRIBUTION = Separator('>>>>>>>> Absorbed dose distribution.')
